@@ -1,5 +1,5 @@
 """Additional checks and filters for problems."""
-
+from functools import wraps
 from io import StringIO
 from typing import Any, Callable, TypeVar
 from unittest.mock import patch
@@ -12,6 +12,7 @@ def with_captured_stdout(
 ) -> Callable[..., tuple[str, Output]]:
     """Run func, returning its stdout and normal return value."""
 
+    @wraps(func)
     def inner(*args: Any, **kwargs: Any) -> tuple[str, Any]:
         with patch("sys.stdout", new_callable=StringIO) as stdout:
             func_out = func(*args, **kwargs)
